# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import Command
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.develop import develop as _develop
from setuptools.command.sdist import sdist as _sdist

from distutils import log

from string import Template
from os import path, makedirs
from datetime import datetime
from urllib.request import urlretrieve
from shutil import unpack_archive, rmtree, copytree

import sys

DEFAULT_COGMENT_API_VERSION = "latest"


class RetrieveCogmentApi(Command):
    description = 'retrieve the cogment api'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        package_dir = path.abspath(self.distribution.package_dir[""])

        from yaml import load

        cogment_api_version = DEFAULT_COGMENT_API_VERSION
        cogment_api_path = None
        # In a lot of context the logs are not displayed,
        # putting everything in a string to be able to create a file output
        retrieve_cogment_api_logs = ""

        fh = open(path.join(package_dir, ".cogment-api.yaml"), "r")
        cogment_api_cfg = load(fh)

        cogment_api_version = cogment_api_cfg.get(
            "cogment_api_version", DEFAULT_COGMENT_API_VERSION
        )
        cogment_api_path = cogment_api_cfg.get("cogment_api_path")

        protos_dir = path.join(package_dir, "cogment/api")

        # honor the --dry-run flag
        if not self.dry_run:
            rmtree(protos_dir, ignore_errors=True)
            makedirs(protos_dir, exist_ok=True)

        if cogment_api_path:
            cogment_api_absolute_path = path.normpath(
                path.join(package_dir, cogment_api_path)
            )
            # honor the --dry-run flag
            if not self.dry_run:
                copytree(cogment_api_absolute_path, protos_dir, dirs_exist_ok=True)
            retrieve_cogment_api_logs += (
                f"cogment api copied from '{cogment_api_absolute_path}'\n"
            )
        else:
            url = "https://cogment.github.io/cogment-api/{cogment_api_version}/cogment-api-{cogment_api_version}.tar.gz".format(
                cogment_api_version=cogment_api_version
            )

            # Download the archive to a temp file, will fail if the version doesn't exist
            # or if the network is down thus keeping what was alredy downloaded.
            temp_local_filename, _ = urlretrieve(url)

            # honor the --dry-run flag
            if not self.dry_run:
                rmtree(protos_dir, ignore_errors=True)

                makedirs(protos_dir, exist_ok=True)

                unpack_archive(temp_local_filename, protos_dir, "gztar")

            retrieve_cogment_api_logs += (
                f"cogment api version '{cogment_api_version}', downloaded from '{url}'\n"
            )

        log.info(retrieve_cogment_api_logs)
        with open(path.join(protos_dir, "cogment-api-retrieval.log"), "wb") as fh:
            fh.write(retrieve_cogment_api_logs.encode("utf-8"))


class BuildCogmentApiProtos(Command):
    description = 'build the cogment api protos'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        package_dir = path.abspath(self.distribution.package_dir[""])

        from grpc.tools import command

        log.info(f"generating python implementation of the proto files in '{package_dir}'")
        # honor the --dry-run flag
        if not self.dry_run:
            command.build_package_protos(package_dir)

class GenerateVersionDotPy(Command):
    description = 'generate cogment/version/py'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        version = self.distribution.get_version()
        package_dir = path.abspath(self.distribution.package_dir[""])
        template_filename = path.join(package_dir, "cogment/version.py.in")
        filename = path.join(package_dir, "cogment/version.py")

        log.info(f"generating version specification file at '{filename}'")

        with open(template_filename, "r") as fh:
            template = Template(fh.read())

        data = template.substitute({"version": version})

        # honor the --dry-run flag
        if not self.dry_run:
            with open(filename, "wb") as fh:
                data = data.encode("utf-8")
                fh.write(data)
                fh.close()

class develop(_develop):
    def run(self):
        self.run_command('retrieve_cogment_api')
        self.run_command('build_cogment_api_protos')
        self.run_command('generate_version_dot_py')
        super().run()

# Unused at the moment
# class sdist(_sdist):
#     def run(self):
#         self.run_command('retrieve_cogment_api')
#         self.run_command('generate_version_dot_py')
#         super().run()


# class build_py(_build_py):
#     def run(self):
#         self.run_command('build_cogment_api_protos')
#         super().run()


# This function accepts a dict containing the keyword arguments that will be passed to setuptools.setup.
# It can then be modified in place as appropriate for your build process.
def build(setup_kwargs):
    setup_kwargs.update(
        {
            "package_dir": {"": "."},
            "package_data": {
                'cogment.api': ['*.proto', '*.py'],
            },
            "setup_requires": ["grpcio-tools>=1.19"],
            "cmdclass": {
                'retrieve_cogment_api': RetrieveCogmentApi,
                'generate_version_dot_py': GenerateVersionDotPy,
                'build_cogment_api_protos': BuildCogmentApiProtos,
                # Executed when the package is installed as a dependency, i.e. when someone does something like pip install cogment.
                # TODO It appears it doesn't work, I can't find a working example on the internet and there is next to no documentation.
                #"build_py": build_py,
                # Executed when the package is installed for development, using `poetry install`.
                "develop": develop,
                # The 'sdist' command should be used to generate the source distribution of the package.
                # However poetry doesn't use it... this shouldn't be a problem if `poetry install` is executed
                # before.
                #"sdist": sdist,
            },
        }
    )



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is the main import module for the cogment (https://gitlab.com/cogment/cogment) AI
framework python SDK by AI Redefined Inc.
"""

from cogment.version import __version__

from cogment.context import Context, Endpoint, ServedEndpoint
from cogment.session import EventType
from cogment.control import TrialState

# Necessary because of cogment CLI "cog_settings.py" generated code
from cogment.actor import ActorClass, ActorClassList
import cogment.delta_encoding



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
from abc import ABC
from cogment.session import Session, EventType

import cogment.api.orchestrator_pb2 as orchestrator_api
import cogment.api.common_pb2 as common_api


class ActorClass:
    def __init__(
        self,
        name,
        config_type,
        action_space,
        observation_space,
        observation_delta,
        observation_delta_apply_fn,
    ):
        self.name = name
        self.config_type = config_type
        self.action_space = action_space
        self.observation_space = observation_space
        self.observation_delta = observation_delta
        self.observation_delta_apply_fn = observation_delta_apply_fn


class ActorClassList:
    def __init__(self, *args):
        self._actor_classes_list = list(args)

        for a_c in args:
            setattr(self, a_c.name, a_c)

    def __iter__(self):
        return iter(self._actor_classes_list)

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def get_class_by_index(self, index):
        return self._actor_classes_list[index]


class ActorSession(Session):
    """This represents an actor being performed locally."""

    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(trial, name, impl, impl_name)
        self.class_name = actor_class.name
        self.config = config

        self._actor_class = actor_class
        self._action_queue = asyncio.Queue()
        self._latest_observation = None

        self.__obs_future = None

    def start(self):
        self._start()

    def do_action(self, action):
        if self._event_queue is None:
            logging.warning(f"Cannot send action until actor [{self.name}] is started.")
            return
        self._action_queue.put_nowait(action)

    async def _retrieve_action(self):
        try:
            action = await self._action_queue.get()
            self._action_queue.task_done()
            return action

        except RuntimeError as exc:
            # Unfortunatelty asyncio returns a standard RuntimeError in this case
            if exc.args[0] != "Event loop is closed":
                raise
            else:
                logging.debug(f"Normal exception on retrieving action at close: [{exc}]")

        return None


class _ServedActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(impl, actor_class, trial, name, impl_name, config)

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        assert self._trial is not None
        self._trial.add_reward(value, confidence, to, tick_id, user_data)

    def send_message(self, payload, to, to_environment=False):
        assert self._trial is not None
        self._trial.send_message(payload, to, to_environment)


class _ClientActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config, actor_stub):
        super().__init__(impl, actor_class, trial, name, impl_name, config)
        self._actor_sub = actor_stub

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        request = orchestrator_api.TrialRewardRequest()
        for dest_actor in self._trial.get_actors(pattern_list=to):
            reward = common_api.Reward(receiver_name=dest_actor.name, tick_id=-1, value=value)
            reward_source = common_api.RewardSource(sender_name=self.name, value=value, confidence=confidence)
            if user_data is not None:
                reward_source.user_data.Pack(user_data)
            reward.sources.append(reward_source)
            request.rewards.append(reward)
            metadata = (("trial-id", self.get_trial_id()), ("actor-name", self.name))
            self._actor_sub.SendReward(request=request, metadata=metadata)

    def send_message(self, payload, to, to_environment=False):
        message_req = orchestrator_api.TrialMessageRequest()
        for dest_actor in self._trial.get_actors(pattern_list=to):
            message = common_api.Message(tick_id=-1, receiver_name=dest_actor.name)
            if payload is not None:
                message.payload.Pack(payload)
            message_req.messages.append(message)
            metadata = (("trial-id", self.get_trial_id()), ("actor-name", self.name))
            self._actor_sub.SendMessage(request=message_req, metadata=metadata)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from enum import Enum
import logging
import traceback
import asyncio


class ActorInfo:
    def __init__(self, name, class_name):
        self.actor_name = name
        self.actor_class_name = class_name

    def __str__(self):
        result = f"ActorInfo: actor name = {self.actor_name}, actor class name = {self.actor_class_name}"
        return result


class RecvObservation:
    def __init__(self, obs, snapshot):
        self.tick_id = obs.tick_id
        self.timestamp = obs.timestamp
        self.snapshot = snapshot

    def __str__(self):
        result = f"RecvObservation: tick_id = {self.tick_id}, timestamp = {self.timestamp}"
        result += f", snapshot = {self.snapshot}"
        return result


class RecvAction:
    def __init__(self, actor_index, action):
        self.actor_index = actor_index
        self.action = action

    def __str__(self):
        result = f"RecvAction: actor index = {self.actor_index}"
        result += f", action = {self.action}"
        return result


class RecvRewardSource:
    def __init__(self, src):
        self.value = src.value
        self.confidence = src.confidence
        self.sender_name = src.sender_name
        self.user_data = src.user_data

    def __str__(self):
        result = f"RecvRewardSource: value = {self.value}, confidence = {self.confidence}"
        result += f", sender name = {self.sender_name}, user data = {self.user_data}"
        return result


class RecvReward:
    def __init__(self, reward):
        self.tick_id = reward.tick_id
        self.value = reward.value
        self._sources = reward.sources

    def get_nb_sources(self):
        return len(self._sources)

    def all_sources(self):
        assert self._sources
        for src in self._sources:
            yield RecvRewardSource(src)

    def __str__(self):
        result = f"RecvReward: tick_id = {self.tick_id}, value = {self.value}"
        result += f", sources = {self._sources}"
        return result


class RecvMessage:
    def __init__(self, message):
        self.tick_id = message.tick_id
        self.sender_name = message.sender_name
        self.payload = message.payload

    def __str__(self):
        result = f"RecvMessage: tick_id = {self.tick_id}, sender_name = {self.sender_name}"
        result += f", payload = {self.payload}"
        return result


class EventType(Enum):
    NONE = 0
    ACTIVE = 1
    ENDING = 2
    FINAL = 3


class RecvEvent:
    def __init__(self, etype):
        self.type = etype
        self.observation = None
        self.actions = []
        self.rewards = []
        self.messages = []

    def __str__(self):
        result = f"RecvEvent: type = {self.type}"
        if self.observation:
            result += f", {{{self.observation}}}"
        for act in self.actions:
            result += f", {{{act}}}"
        for rew in self.rewards:
            result += f", {{{rew}}}"
        for msg in self.messages:
            result += f", {{{msg}}}"
        return result


class Session(ABC):
    def __init__(self, trial, name, impl, impl_name):
        self._trial = trial
        self.name = name
        self.impl_name = impl_name
        self._impl = impl
        self._event_queue = None  # Also used to know if we started
        self._last_event_delivered = False
        self._last_event_received = False
        self._task = None  # Task used to call _run()

        # Pre-compute since it will be used regularly
        self._active_actors = [ActorInfo(actor.name, actor.actor_class.name) for actor in trial.actors]

    def _start(self):
        if self._event_queue is not None:
            logging.warning(f"Cannot start [{self.name}] more than once.")
            return
        if self._trial.over:
            logging.warning(f"Cannot start [{self.name}] because the trial has ended.")
            return

        self._event_queue = asyncio.Queue()

    async def _run(self):
        try:
            await self._impl(self)

        except asyncio.CancelledError:
            logging.debug(f"[{self.name}] implementation coroutine cancelled")

        except Exception:
            logging.error(f"An exception occured in user implementation of [{self.name}]:\n{traceback.format_exc()}")
            raise

    def _new_event(self, event):
        if self._event_queue is None:
            logging.warning(f"[{self.name}] received an event before session was started.")
            return
        if self._last_event_received:
            logging.debug(f"Event received after final event: [{event}]")
            return

        if event is not None and event.type == EventType.FINAL:
            self._last_event_received = True

        self._event_queue.put_nowait(event)

    async def event_loop(self):
        if self._event_queue is None:
            logging.warning(f"Event loop is not enabled until the [{self.name}] is started.")
            return
        if self._trial.over:
            logging.info(f"No more events for [{self.name}] because the trial has ended.")
            return

        loop_active = not self._last_event_delivered
        while loop_active:
            try:
                event = await self._event_queue.get()

            except asyncio.CancelledError:
                logging.debug(f"[{self.name}] coroutine cancelled while waiting for an event")
                break

            self._last_event_delivered = (event.type == EventType.FINAL)
            keep_looping = yield event
            self._event_queue.task_done()
            loop_active = (keep_looping is None or bool(keep_looping)) and not self._last_event_delivered
            if not loop_active:
                if self._last_event_delivered:
                    logging.debug(f"Last event delivered, exiting [{self.name}] event loop")
                else:
                    logging.debug(f"End of event loop for [{self.name}] requested by user")

        logging.debug(f"Exiting [{self.name}] event loop generator")

    def get_trial_id(self):
        assert self._trial is not None
        return self._trial.id

    def get_tick_id(self):
        assert self._trial is not None
        return self._trial.tick_id

    def is_trial_over(self):
        assert self._trial is not None
        return self._trial.over

    def get_active_actors(self):
        assert self._trial is not None
        return self._active_actors

    @abstractmethod
    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        pass

    @abstractmethod
    def send_message(self, payload, to, to_environment=False):
        pass



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from cogment.errors import InvalidParamsError


class PrehookSession(ABC):
    """This represents the session for a prehook for a trial"""

    def __init__(self, params, trial):
        self._params = params
        self._trial = trial

        self.trial_config = params["trial_config"]
        self.trial_max_steps = params["max_steps"]
        self.trial_max_inactivity = params["max_inactivity"]

        self.environment_config = params["environment"]["config"]
        self.environment_endpoint = params["environment"]["endpoint"]

        self.actors = params["actors"]

    @abstractmethod
    def _recode(self):
        pass

    def get_trial_id(self):
        return self._trial.id

    def validate(self):
        self._recode()


class _ServedPrehookSession(PrehookSession):
    def __init__(self, params, trial):
        super().__init__(params, trial)

    def _recode(self):
        for act in self.actors:
            if "name" not in act or \
               "actor_class" not in act or \
               "endpoint" not in act or \
               "implementation" not in act or \
               "config" not in act:
                raise InvalidParamsError(f"incomplete actor: {act}")

        self._params["actors"] = self.actors

        self._params["trial_config"] = self.trial_config
        self._params["max_steps"] = self.trial_max_steps
        self._params["max_inactivity"] = self.trial_max_inactivity

        self._params["environment"]["config"] = self.environment_config
        self._params["environment"]["endpoint"] = self.environment_endpoint



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import atexit
import signal
import threading
import logging

# from cogment.hooks_service import HooksService, TrialHooks
# from cogment.agent_service import AgentService, Agent
# from cogment.env_service import EnvService, Environment
from cogment.utils import list_versions

import cogment.api.hooks_pb2_grpc as grpc_hooks_api
import cogment.api.environment_pb2_grpc as grpc_env_api
import cogment.api.agent_pb2_grpc as grpc_agent_api

from cogment.api.environment_pb2 import _ENVIRONMENTENDPOINT as env_descriptor
from cogment.api.agent_pb2 import _AGENTENDPOINT as agent_descriptor
from cogment.api.hooks_pb2 import _TRIALHOOKS as hooks_descriptor

# from cogment.errors import ConfigError
from cogment.reloader import run_with_reloader

from grpc_reflection.v1alpha import reflection
from concurrent.futures import ThreadPoolExecutor
import grpc

from distutils.util import strtobool

ENABLE_REFLECTION_VAR_NAME = 'COGMENT_GRPC_REFLECTION'
ENABLE_GRPC_SERVER_AUTORELOAD_VAR_NAME = 'COGMENT_GRPC_SERVER_RELOAD'
DEFAULT_PORT = 9000
MAX_WORKERS = 10


# A Grpc endpoint serving a cogment service
class GrpcServer:
    def __init__(self, service_type, settings, port=DEFAULT_PORT):
        self.__exit_handler = None
        self._port = port
        self._grpc_server = grpc.server(ThreadPoolExecutor(
            max_workers=MAX_WORKERS))

        self._service_types = []

        if isinstance(service_type, list):
            for s in service_type:
                self._add_service(s, settings)
        else:
            self._add_service(service_type, settings)

        # Enable grpc reflection if requested
        if strtobool(os.getenv(ENABLE_REFLECTION_VAR_NAME, 'false')):
            service_names = [s.full_name for s in self._service_types] + [reflection.SERVICE_NAME]
            reflection.enable_server_reflection(service_names, self._grpc_server)

        self.__auto_reload = strtobool(os.getenv(ENABLE_GRPC_SERVER_AUTORELOAD_VAR_NAME, 'false'))

        # Give the server a chance to properly shutdown when running in auto_reload mode. Note, we cannot do this
        # by capturing signals as the server is not running in the main thread, when auto_reload is on.
        if self.__auto_reload:
            self.__exit_handler = self.stop
            atexit.register(self.__exit_handler)

        # This check is required because when auto_reload is requested is on the grpc_server won't be launched from
        # the main thread so attempting to capture any of the signals below will just raise an exception.
        if threading.current_thread() is threading.main_thread():
            for sig in ('TERM', 'HUP', 'INT'):
                signal.signal(getattr(signal, 'SIG' + sig), self.stop)

    def _add_service(self, service_type, settings):
        """Adds a service to the grpc server.
           This only works before the server is actually started."""
        logging.info(f"Versions for {service_type.__name__}:")
        for v in list_versions(service_type).versions:
            logging.info(f'  {v.name}: {v.version}')

        # Register service
        if issubclass(service_type, Agent):
            self._service_types.append(agent_descriptor)
            grpc_agent_api.add_AgentEndpointServicer_to_server(
                AgentService(service_type, settings), self._grpc_server)
        elif issubclass(service_type, Environment):
            self._service_types.append(env_descriptor)
            grpc_env_api.add_EnvironmentEndpointServicer_to_server(
                EnvService(service_type, settings), self._grpc_server)
        elif issubclass(service_type, TrialHooks):
            self._service_types.append(hooks_descriptor)
            grpc_hooks_api.add_TrialHooksServicer_to_server(
                HooksService(service_type, settings), self._grpc_server)
        else:
            raise ConfigError('Invalid service type')

    def __run(self):
        self._grpc_server.add_insecure_port(f'[::]:{self._port}')
        self._grpc_server.start()

        for s in self._service_types:
            logging.info(f"{s.full_name} service listening on port {self._port}")

        self._grpc_server.wait_for_termination()

    def serve(self):
        if self.__auto_reload:
            run_with_reloader(self.__run)
        else:
            self.__run()

    def stop(self, *args):
        if self.__exit_handler:
            atexit.unregister(self.__exit_handler)
            self.__exit_handler = None

        self._grpc_server.stop(0).wait()



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.common_pb2 as common_api
import cogment.api.orchestrator_pb2 as orchestrator_api
from cogment.session import ActorInfo

import asyncio
import grpc
import grpc.experimental.aio
from enum import Enum
import logging
import traceback


class TrialState(Enum):
    UNKNOWN = common_api.TrialState.UNKNOWN
    INITIALIZING = common_api.TrialState.INITIALIZING
    PENDING = common_api.TrialState.PENDING
    RUNNING = common_api.TrialState.RUNNING
    TERMINATING = common_api.TrialState.TERMINATING
    ENDED = common_api.TrialState.ENDED


class TrialInfo:
    def __init__(self, trial_id):
        self.trial_id = trial_id
        self.state = TrialState.UNKNOWN
        self.tick_id = None
        self.duration = None

    def __str__(self):
        result = f"TrialInfo: trial_id = {self.trial_id}, state = {self.state}"
        result += f", tick_id = {self.tick_id}, duration = {self.duration}"
        return result


class Controller:
    def __init__(self, stub, user_id):
        self._lifecycle_stub = stub
        self._user_id = user_id

        # TODO: This should be removed once we have a stable release
        logging.basicConfig(level=logging.DEBUG)

    async def get_actors(self, trial_id):
        req = orchestrator_api.TrialInfoRequest()
        req.get_actor_list = True
        metadata = [("trial-id", trial_id)]
        rep = await self._lifecycle_stub.GetTrialInfo(request=req, metadata=metadata)

        assert len(rep.trial) == 1
        result = [ActorInfo(actor.name, actor.actor_class) for actor in rep.trial[0].actors_in_trial]
        return result

    async def start_trial(self, trial_config=None):
        req = orchestrator_api.TrialStartRequest()
        req.user_id = self._user_id
        if trial_config is not None:
            req.config.content = trial_config.SerializeToString()

        logging.debug(f"Requesting start of a trial with [{req}] ...")
        rep = await self._lifecycle_stub.StartTrial(req)
        logging.debug(f"Trial [{rep.trial_id}] started")

        return rep.trial_id

    async def terminate_trial(self, trial_id):
        req = orchestrator_api.TerminateTrialRequest()
        metadata = [("trial-id", trial_id)]
        logging.debug(f"Requesting end of trial [{trial_id}] ...")
        await self._lifecycle_stub.TerminateTrial(request=req, metadata=metadata)
        logging.debug(f"End of trial request accepted for {trial_id}")

    async def get_trial_info(self, trial_id):
        req = orchestrator_api.TrialInfoRequest()
        if trial_id is not None:
            metadata = [("trial-id", trial_id)]
            rep = await self._lifecycle_stub.GetTrialInfo(request=req, metadata=metadata)
        else:
            rep = await self._lifecycle_stub.GetTrialInfo(request=req)

        result = []
        for reply in rep.trial:
            info_ex = TrialInfo(reply.trial_id)
            info_ex.state = TrialState(reply.state)
            info_ex.tick_id = reply.tick_id
            info_ex.duration = reply.trial_duration

            result.append(info_ex)

        return result

    async def watch_trials(self, trial_state_filters=[]):
        request = orchestrator_api.TrialListRequest()
        for fil in trial_state_filters:
            request.filter.append(fil.value)

        reply_itor = self._lifecycle_stub.WatchTrials(request=request)
        if not reply_itor:
            raise Exception(f"'watch_trials' failed to connect")

        try:
            async for reply in reply_itor:
                info = TrialInfo(reply.trial_id)
                info.state = TrialState(reply.state)
                keep_looping = yield info
                if keep_looping is not None and not bool(keep_looping):
                    break

        except grpc.experimental.aio._call.AioRpcError as exc:
            logging.debug(f"gRPC Error details: [{exc.debug_error_string()}]")
            if exc.code() == grpc.StatusCode.UNAVAILABLE:
                logging.error(f"Orchestrator communication lost: [{exc.details()}]")
            else:
                logging.error(f"{traceback.format_exc()}")
                raise

        except GeneratorExit:
            raise

        except asyncio.CancelledError:
            logging.debug(f"watch_trial coroutine cancelled while waiting for trial info")

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Callable, Awaitable, Dict, List, Any
from types import ModuleType
import os
import asyncio
from types import SimpleNamespace
import grpc
import grpc.experimental.aio
from prometheus_client import start_http_server, CollectorRegistry

from cogment.actor import ActorSession, ActorClass
from cogment.environment import EnvironmentSession
from cogment.prehook import PrehookSession
from cogment.datalog import DatalogSession
from cogment.control import Controller

# Controller
import cogment.api.orchestrator_pb2_grpc as grpc_api

# Agent
from cogment.agent_service import AgentServicer
from cogment.api.agent_pb2 import _AGENTENDPOINT as agent_endpoint_descriptor
from cogment.api.agent_pb2_grpc import add_AgentEndpointServicer_to_server

# Client
from cogment.client_service import ClientServicer

# Environment
from cogment.env_service import EnvironmentServicer
from cogment.api.environment_pb2_grpc import add_EnvironmentEndpointServicer_to_server

# Prehook
from cogment.hooks_service import PrehookServicer
from cogment.api.hooks_pb2_grpc import add_TrialHooksServicer_to_server

# Log Exporter
from cogment.log_exporter_service import LogExporterService
from cogment.api.datalog_pb2_grpc import add_LogExporterServicer_to_server


DEFAULT_MAX_WORKERS = 1
ENABLE_REFLECTION_VAR_NAME = "COGMENT_GRPC_REFLECTION"
DEFAULT_ENABLE_REFLECTION = os.getenv(ENABLE_REFLECTION_VAR_NAME, "false")
DEFAULT_PROMETHEUS_PORT = 8000


def _add_actor_service(grpc_server, impls, service_names, cog_settings, prometheus_registry):
    servicer = AgentServicer(impls, cog_settings, prometheus_registry)
    add_AgentEndpointServicer_to_server(servicer, grpc_server)
    service_names.append(agent_endpoint_descriptor.full_name)


def _add_env_service(grpc_server, impls, cog_settings, prometheus_registry):
    servicer = EnvironmentServicer(impls, cog_settings, prometheus_registry)
    add_EnvironmentEndpointServicer_to_server(servicer, grpc_server)


def _add_prehook_service(grpc_server, impls, cog_settings, prometheus_registry):
    servicer = PrehookServicer(impls, cog_settings, prometheus_registry)
    add_TrialHooksServicer_to_server(servicer, grpc_server)


def _add_datalog_service(grpc_server, impl, cog_settings):
    servicer = LogExporterService(impl, cog_settings)
    add_LogExporterServicer_to_server(servicer, grpc_server)


class Endpoint:

    def __init__(self, url: str):
        self.url = url
        self.private_key = None
        self.root_certificates = None
        self.certificate_chain = None

    def set_from_files(self, private_key_file=None, root_certificates_file=None, certificate_chain_file=None):
        try:
            if private_key_file:
                with open(private_key_file) as fl:
                    self.private_key = fl.read()
                if not self.private_key:
                    self.private_key = None

            if root_certificates_file:
                with open(root_certificates_file) as fl:
                    self.root_certificates = fl.read()
                if not self.root_certificates:
                    self.root_certificates = None

            if certificate_chain_file:
                with open(certificate_chain_file) as fl:
                    self.certificate_chain = fl.read()
                if not self.certificate_chain:
                    self.certificate_chain = None

        except Exception:
            logging.error(f"Failed loading file from CWD: {os.getcwd()}")
            raise


class ServedEndpoint:

    def __init__(self, port: int):
        self.port = port
        self.private_key_certificate_chain_pairs = None
        self.root_certificates = None

    # def set_from_files(private_key_certificate_chain_pairs_file=None, root_certificates_file=None):
    # TODO: This function would need to parse the PEM encoded `private_key_certificate_chain_pairs_file`
    #       to create the list of tuples required


class Context:
    def __init__(self, user_id: str, cog_settings: ModuleType, asyncio_loop=None):
        self._user_id = user_id
        self.__actor_impls: Dict[str, SimpleNamespace] = {}
        self.__env_impls: Dict[str, SimpleNamespace] = {}
        self.__prehook_impls: List[Callable[[PrehookSession], Awaitable[None]]] = []
        self.__datalog_impl: Callable[[DatalogSession], Awaitable[None]] = None
        self._grpc_server = None  # type: Any
        self._prometheus_registry = CollectorRegistry()
        self.__cog_settings = cog_settings

        if asyncio_loop is None:
            # Make sure we are running in a asyncio.Task.  Even if technically this init does not need
            # to be in a Task, the whole of the SDK expects to be running in Tasks of the same event loop.
            try:
                self.asyncio_loop = asyncio.get_running_loop()
            except RuntimeError:
                pass
            if self.asyncio_loop is None:
                raise RuntimeError("The Cogment SDK requires running in a Python.asyncio Task")
        else:
            self.asyncio_loop = asyncio_loop

    def register_actor(self,
                       impl: Callable[[ActorSession], Awaitable[None]],
                       impl_name: str,
                       actor_classes: List[str] = []):
        assert impl_name not in self.__actor_impls
        assert self._grpc_server is None

        # TODO: Add test to make sure that if actor is added when _grps_server is not None, then
        # it is a client actor (i.e. only client actors can be started after `serve_all_registered` was called)
        self.__actor_impls[impl_name] = SimpleNamespace(impl=impl, actor_classes=actor_classes)

    def register_environment(self,
                             impl: Callable[[EnvironmentSession], Awaitable[None]],
                             impl_name: str = "default"):
        assert impl_name not in self.__env_impls
        assert self._grpc_server is None

        self.__env_impls[impl_name] = SimpleNamespace(impl=impl)

    def register_pre_trial_hook(self,
                                impl: Callable[[PrehookSession], Awaitable[None]]):
        assert self._grpc_server is None

        self.__prehook_impls.append(impl)

    def register_datalog(self,
                         impl: Callable[[DatalogSession], Awaitable[None]]):
        assert self._grpc_server is None
        assert self.__datalog_impl is None

        self.__datalog_impl = impl

    async def serve_all_registered(self, served_endpoint: ServedEndpoint,
                                   prometheus_port: int = DEFAULT_PROMETHEUS_PORT):
        if (len(self.__actor_impls) == 0 and
                len(self.__env_impls) == 0 and
                len(self.__prehook_impls) == 0 and
                self.__datalog_impl is None):
            raise RuntimeError("Nothing registered to serve!")
        if self._grpc_server is not None:
            raise RuntimeError("Cannot serve the same components twice")

        self._grpc_server = grpc.experimental.aio.server()

        service_names: List[str] = []
        if self.__actor_impls:
            _add_actor_service(
                self._grpc_server,
                self.__actor_impls,
                service_names,
                self.__cog_settings,
                self._prometheus_registry
            )

        if self.__env_impls:
            _add_env_service(
                self._grpc_server,
                self.__env_impls,
                self.__cog_settings,
                self._prometheus_registry
            )

        if self.__prehook_impls:
            _add_prehook_service(
                self._grpc_server,
                self.__prehook_impls,
                self.__cog_settings,
                self._prometheus_registry
            )

        if self.__datalog_impl is not None:
            _add_datalog_service(
                self._grpc_server,
                self.__datalog_impl,
                self.__cog_settings
            )

        start_http_server(prometheus_port, "", self._prometheus_registry)

        address = f"[::]:{served_endpoint.port}"
        if served_endpoint.private_key_certificate_chain_pairs is None:
            self._grpc_server.add_insecure_port(address)
        else:
            if served_endpoint.root_certificates:
                require_client_auth = True
                root = bytes(served_endpoint.root_certificates, "utf-8")
            else:
                require_client_auth = False
                root = None
            certs = []
            for (key, chain) in served_endpoint.private_key_certificate_chain_pairs:
                certs.append((bytes(key, "utf-8"), bytes(chain, "utf-8")))
            if not certs:
                certs = None

            creds = grpc.ssl_server_credentials(certs, root, require_client_auth)
            self._grpc_server.add_secure_port(address, creds)

        await self._grpc_server.start()
        await self._grpc_server.wait_for_termination()
        logging.debug(f"Context gRPC server at port [{served_endpoint.port}] for user [{self._user_id}] exited")

    def _get_control_stub(self, endpoint):
        if endpoint.private_key is None:
            channel = grpc.experimental.aio.insecure_channel(endpoint.url)
        else:
            if endpoint.root_certificates:
                root = bytes(endpoint.root_certificates, "utf-8")
            else:
                root = None
            if endpoint.private_key:
                key = bytes(endpoint.private_key, "utf-8")
            else:
                key = None
            if endpoint.certificate_chain:
                certs = bytes(endpoint.certificate_chain, "utf-8")
            else:
                certs = None
            creds = grpc.ssl_channel_credentials(root, key, certs)
            channel = grpc.experimental.aio.secure_channel(endpoint.url, creds)

        return grpc_api.TrialLifecycleStub(channel)

    def get_controller(self, endpoint: Endpoint):
        stub = self._get_control_stub(endpoint)
        return Controller(stub, self._user_id)

    async def join_trial(self, trial_id, endpoint: Endpoint, impl_name, actor_name=None):
        actor_impl = self.__actor_impls.get(impl_name)
        if actor_impl is None:
            raise Exception(f"Unknown actor impl [{impl_name}].  Was it registered?")

        servicer = ClientServicer(self.__cog_settings, endpoint)
        await servicer.run(trial_id, actor_impl.impl, impl_name, actor_impl.actor_classes, actor_name)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import traceback
from abc import ABC, abstractmethod


class DatalogSession(ABC):
    """This represents a datalogger working locally."""

    def __init__(self, impl, trial_id, trial_params, raw_trial_params):
        self.trial_id = trial_id
        self.trial_params = trial_params
        self.raw_trial_params = raw_trial_params

        self._task = None
        self.__impl = impl
        self.__queue = None

    def start(self):
        self.__queue = asyncio.Queue()

    async def get_all_samples(self):
        if self.__queue is not None:
            while True:
                try:
                    sample = await self.__queue.get()
                    if sample is None:
                        break
                    keep_looping = yield sample
                    if keep_looping is not None and not bool(keep_looping):
                        break

                    self.__queue.task_done()

                except GeneratorExit:
                    raise

                except asyncio.CancelledError:
                    logging.debug("Datalog coroutine cancelled while waiting for a sample.")
                    break

    def _new_sample(self, sample):
        if self.__queue is not None:
            self.__queue.put_nowait(sample)
        else:
            logging.warning("Datalog received a sample that it was unable to handle.")

    async def _run(self):
        try:
            await self.__impl(self)
        except Exception:
            logging.error(f"An exception occured in user datalog implementation:\n{traceback.format_exc()}")
            raise


class _ServedDatalogSession(DatalogSession):
    def __init__(self, impl, trial_id, trial_params, raw_trial_params):
        super().__init__(impl, trial_id, trial_params, raw_trial_params)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/orchestrator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import environment_pb2 as cogment_dot_api_dot_environment__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/orchestrator.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1e\x63ogment/api/orchestrator.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\x1a\x1d\x63ogment/api/environment.proto\"J\n\x11TrialStartRequest\x12$\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x14.cogment.TrialConfig\x12\x0f\n\x07user_id\x18\x02 \x01(\t\"Q\n\x0fTrialStartReply\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12,\n\x0f\x61\x63tors_in_trial\x18\x02 \x03(\x0b\x32\x13.cogment.TrialActor\"\x17\n\x15TerminateTrialRequest\"\x15\n\x13TerminateTrialReply\"J\n\x10TrialInfoRequest\x12\x1e\n\x16get_latest_observation\x18\x01 \x01(\x08\x12\x16\n\x0eget_actor_list\x18\x02 \x01(\x08\"\xcd\x01\n\tTrialInfo\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12\"\n\x05state\x18\x02 \x01(\x0e\x32\x13.cogment.TrialState\x12\x0f\n\x07tick_id\x18\x03 \x01(\x04\x12\x16\n\x0etrial_duration\x18\x04 \x01(\x06\x12\x33\n\x12latest_observation\x18\x05 \x01(\x0b\x32\x17.cogment.ObservationSet\x12,\n\x0f\x61\x63tors_in_trial\x18\x06 \x03(\x0b\x32\x13.cogment.TrialActor\"3\n\x0eTrialInfoReply\x12!\n\x05trial\x18\x01 \x03(\x0b\x32\x12.cogment.TrialInfo\"7\n\x10TrialListRequest\x12#\n\x06\x66ilter\x18\x01 \x03(\x0e\x32\x13.cogment.TrialState\"F\n\x0eTrialListEntry\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12\"\n\x05state\x18\x02 \x01(\x0e\x32\x13.cogment.TrialState\"c\n\x10TrialJoinRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12\x15\n\x0b\x61\x63tor_class\x18\x02 \x01(\tH\x00\x12\x14\n\nactor_name\x18\x03 \x01(\tH\x00\x42\x10\n\x0eslot_selection\"\x8a\x01\n\x0eTrialJoinReply\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x10\n\x08trial_id\x18\x02 \x01(\t\x12$\n\x06\x63onfig\x18\x03 \x01(\x0b\x32\x14.cogment.ActorConfig\x12,\n\x0f\x61\x63tors_in_trial\x18\x04 \x03(\x0b\x32\x13.cogment.TrialActor\"\x17\n\x15TrialHeartbeatRequest\"\x15\n\x13TrialHeartbeatReply\"5\n\x12TrialActionRequest\x12\x1f\n\x06\x61\x63tion\x18\x01 \x01(\x0b\x32\x0f.cogment.Action\"N\n\x10TrialActionReply\x12&\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32\x18.cogment.ActorPeriodData\x12\x12\n\nfinal_data\x18\x02 \x01(\x08\"6\n\x12TrialRewardRequest\x12 \n\x07rewards\x18\x01 \x03(\x0b\x32\x0f.cogment.Reward\"\x12\n\x10TrialRewardReply\"9\n\x13TrialMessageRequest\x12\"\n\x08messages\x18\x01 \x03(\x0b\x32\x10.cogment.Message\"\x13\n\x11TrialMessageReply2\xf1\x02\n\x0eTrialLifecycle\x12\x44\n\nStartTrial\x12\x1a.cogment.TrialStartRequest\x1a\x18.cogment.TrialStartReply\"\x00\x12P\n\x0eTerminateTrial\x12\x1e.cogment.TerminateTrialRequest\x1a\x1c.cogment.TerminateTrialReply\"\x00\x12\x44\n\x0cGetTrialInfo\x12\x19.cogment.TrialInfoRequest\x1a\x17.cogment.TrialInfoReply\"\x00\x12\x45\n\x0bWatchTrials\x12\x19.cogment.TrialListRequest\x1a\x17.cogment.TrialListEntry\"\x00\x30\x01\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x32\xba\x03\n\x0b\x43lientActor\x12\x41\n\tJoinTrial\x12\x19.cogment.TrialJoinRequest\x1a\x17.cogment.TrialJoinReply\"\x00\x12L\n\x0c\x41\x63tionStream\x12\x1b.cogment.TrialActionRequest\x1a\x19.cogment.TrialActionReply\"\x00(\x01\x30\x01\x12K\n\tHeartbeat\x12\x1e.cogment.TrialHeartbeatRequest\x1a\x1c.cogment.TrialHeartbeatReply\"\x00\x12\x46\n\nSendReward\x12\x1b.cogment.TrialRewardRequest\x1a\x19.cogment.TrialRewardReply\"\x00\x12I\n\x0bSendMessage\x12\x1c.cogment.TrialMessageRequest\x1a\x1a.cogment.TrialMessageReply\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,cogment_dot_api_dot_environment__pb2.DESCRIPTOR,])




_TRIALSTARTREQUEST = _descriptor.Descriptor(
  name='TrialStartRequest',
  full_name='cogment.TrialStartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.TrialStartRequest.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='cogment.TrialStartRequest.user_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=174,
)


_TRIALSTARTREPLY = _descriptor.Descriptor(
  name='TrialStartReply',
  full_name='cogment.TrialStartReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialStartReply.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.TrialStartReply.actors_in_trial', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=176,
  serialized_end=257,
)


_TERMINATETRIALREQUEST = _descriptor.Descriptor(
  name='TerminateTrialRequest',
  full_name='cogment.TerminateTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=282,
)


_TERMINATETRIALREPLY = _descriptor.Descriptor(
  name='TerminateTrialReply',
  full_name='cogment.TerminateTrialReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=284,
  serialized_end=305,
)


_TRIALINFOREQUEST = _descriptor.Descriptor(
  name='TrialInfoRequest',
  full_name='cogment.TrialInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='get_latest_observation', full_name='cogment.TrialInfoRequest.get_latest_observation', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='get_actor_list', full_name='cogment.TrialInfoRequest.get_actor_list', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=307,
  serialized_end=381,
)


_TRIALINFO = _descriptor.Descriptor(
  name='TrialInfo',
  full_name='cogment.TrialInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialInfo.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='cogment.TrialInfo.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tick_id', full_name='cogment.TrialInfo.tick_id', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_duration', full_name='cogment.TrialInfo.trial_duration', index=3,
      number=4, type=6, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_observation', full_name='cogment.TrialInfo.latest_observation', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.TrialInfo.actors_in_trial', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=384,
  serialized_end=589,
)


_TRIALINFOREPLY = _descriptor.Descriptor(
  name='TrialInfoReply',
  full_name='cogment.TrialInfoReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial', full_name='cogment.TrialInfoReply.trial', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=591,
  serialized_end=642,
)


_TRIALLISTREQUEST = _descriptor.Descriptor(
  name='TrialListRequest',
  full_name='cogment.TrialListRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter', full_name='cogment.TrialListRequest.filter', index=0,
      number=1, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=644,
  serialized_end=699,
)


_TRIALLISTENTRY = _descriptor.Descriptor(
  name='TrialListEntry',
  full_name='cogment.TrialListEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialListEntry.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='cogment.TrialListEntry.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=701,
  serialized_end=771,
)


_TRIALJOINREQUEST = _descriptor.Descriptor(
  name='TrialJoinRequest',
  full_name='cogment.TrialJoinRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialJoinRequest.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actor_class', full_name='cogment.TrialJoinRequest.actor_class', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actor_name', full_name='cogment.TrialJoinRequest.actor_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='slot_selection', full_name='cogment.TrialJoinRequest.slot_selection',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=773,
  serialized_end=872,
)


_TRIALJOINREPLY = _descriptor.Descriptor(
  name='TrialJoinReply',
  full_name='cogment.TrialJoinReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='actor_name', full_name='cogment.TrialJoinReply.actor_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialJoinReply.trial_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.TrialJoinReply.config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.TrialJoinReply.actors_in_trial', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=875,
  serialized_end=1013,
)


_TRIALHEARTBEATREQUEST = _descriptor.Descriptor(
  name='TrialHeartbeatRequest',
  full_name='cogment.TrialHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1015,
  serialized_end=1038,
)


_TRIALHEARTBEATREPLY = _descriptor.Descriptor(
  name='TrialHeartbeatReply',
  full_name='cogment.TrialHeartbeatReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1040,
  serialized_end=1061,
)


_TRIALACTIONREQUEST = _descriptor.Descriptor(
  name='TrialActionRequest',
  full_name='cogment.TrialActionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action', full_name='cogment.TrialActionRequest.action', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1063,
  serialized_end=1116,
)


_TRIALACTIONREPLY = _descriptor.Descriptor(
  name='TrialActionReply',
  full_name='cogment.TrialActionReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='cogment.TrialActionReply.data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='final_data', full_name='cogment.TrialActionReply.final_data', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1118,
  serialized_end=1196,
)


_TRIALREWARDREQUEST = _descriptor.Descriptor(
  name='TrialRewardRequest',
  full_name='cogment.TrialRewardRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='rewards', full_name='cogment.TrialRewardRequest.rewards', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1198,
  serialized_end=1252,
)


_TRIALREWARDREPLY = _descriptor.Descriptor(
  name='TrialRewardReply',
  full_name='cogment.TrialRewardReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1254,
  serialized_end=1272,
)


_TRIALMESSAGEREQUEST = _descriptor.Descriptor(
  name='TrialMessageRequest',
  full_name='cogment.TrialMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.TrialMessageRequest.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1274,
  serialized_end=1331,
)


_TRIALMESSAGEREPLY = _descriptor.Descriptor(
  name='TrialMessageReply',
  full_name='cogment.TrialMessageReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1333,
  serialized_end=1352,
)

_TRIALSTARTREQUEST.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._TRIALCONFIG
_TRIALSTARTREPLY.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_TRIALINFO.fields_by_name['state'].enum_type = cogment_dot_api_dot_common__pb2._TRIALSTATE
_TRIALINFO.fields_by_name['latest_observation'].message_type = cogment_dot_api_dot_environment__pb2._OBSERVATIONSET
_TRIALINFO.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_TRIALINFOREPLY.fields_by_name['trial'].message_type = _TRIALINFO
_TRIALLISTREQUEST.fields_by_name['filter'].enum_type = cogment_dot_api_dot_common__pb2._TRIALSTATE
_TRIALLISTENTRY.fields_by_name['state'].enum_type = cogment_dot_api_dot_common__pb2._TRIALSTATE
_TRIALJOINREQUEST.oneofs_by_name['slot_selection'].fields.append(
  _TRIALJOINREQUEST.fields_by_name['actor_class'])
_TRIALJOINREQUEST.fields_by_name['actor_class'].containing_oneof = _TRIALJOINREQUEST.oneofs_by_name['slot_selection']
_TRIALJOINREQUEST.oneofs_by_name['slot_selection'].fields.append(
  _TRIALJOINREQUEST.fields_by_name['actor_name'])
_TRIALJOINREQUEST.fields_by_name['actor_name'].containing_oneof = _TRIALJOINREQUEST.oneofs_by_name['slot_selection']
_TRIALJOINREPLY.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._ACTORCONFIG
_TRIALJOINREPLY.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_TRIALACTIONREQUEST.fields_by_name['action'].message_type = cogment_dot_api_dot_common__pb2._ACTION
_TRIALACTIONREPLY.fields_by_name['data'].message_type = cogment_dot_api_dot_common__pb2._ACTORPERIODDATA
_TRIALREWARDREQUEST.fields_by_name['rewards'].message_type = cogment_dot_api_dot_common__pb2._REWARD
_TRIALMESSAGEREQUEST.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
DESCRIPTOR.message_types_by_name['TrialStartRequest'] = _TRIALSTARTREQUEST
DESCRIPTOR.message_types_by_name['TrialStartReply'] = _TRIALSTARTREPLY
DESCRIPTOR.message_types_by_name['TerminateTrialRequest'] = _TERMINATETRIALREQUEST
DESCRIPTOR.message_types_by_name['TerminateTrialReply'] = _TERMINATETRIALREPLY
DESCRIPTOR.message_types_by_name['TrialInfoRequest'] = _TRIALINFOREQUEST
DESCRIPTOR.message_types_by_name['TrialInfo'] = _TRIALINFO
DESCRIPTOR.message_types_by_name['TrialInfoReply'] = _TRIALINFOREPLY
DESCRIPTOR.message_types_by_name['TrialListRequest'] = _TRIALLISTREQUEST
DESCRIPTOR.message_types_by_name['TrialListEntry'] = _TRIALLISTENTRY
DESCRIPTOR.message_types_by_name['TrialJoinRequest'] = _TRIALJOINREQUEST
DESCRIPTOR.message_types_by_name['TrialJoinReply'] = _TRIALJOINREPLY
DESCRIPTOR.message_types_by_name['TrialHeartbeatRequest'] = _TRIALHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['TrialHeartbeatReply'] = _TRIALHEARTBEATREPLY
DESCRIPTOR.message_types_by_name['TrialActionRequest'] = _TRIALACTIONREQUEST
DESCRIPTOR.message_types_by_name['TrialActionReply'] = _TRIALACTIONREPLY
DESCRIPTOR.message_types_by_name['TrialRewardRequest'] = _TRIALREWARDREQUEST
DESCRIPTOR.message_types_by_name['TrialRewardReply'] = _TRIALREWARDREPLY
DESCRIPTOR.message_types_by_name['TrialMessageRequest'] = _TRIALMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['TrialMessageReply'] = _TRIALMESSAGEREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TrialStartRequest = _reflection.GeneratedProtocolMessageType('TrialStartRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialStartRequest)
  })
_sym_db.RegisterMessage(TrialStartRequest)

TrialStartReply = _reflection.GeneratedProtocolMessageType('TrialStartReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialStartReply)
  })
_sym_db.RegisterMessage(TrialStartReply)

TerminateTrialRequest = _reflection.GeneratedProtocolMessageType('TerminateTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TerminateTrialRequest)
  })
_sym_db.RegisterMessage(TerminateTrialRequest)

TerminateTrialReply = _reflection.GeneratedProtocolMessageType('TerminateTrialReply', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TerminateTrialReply)
  })
_sym_db.RegisterMessage(TerminateTrialReply)

TrialInfoRequest = _reflection.GeneratedProtocolMessageType('TrialInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfoRequest)
  })
_sym_db.RegisterMessage(TrialInfoRequest)

TrialInfo = _reflection.GeneratedProtocolMessageType('TrialInfo', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFO,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfo)
  })
_sym_db.RegisterMessage(TrialInfo)

TrialInfoReply = _reflection.GeneratedProtocolMessageType('TrialInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfoReply)
  })
_sym_db.RegisterMessage(TrialInfoReply)

TrialListRequest = _reflection.GeneratedProtocolMessageType('TrialListRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALLISTREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialListRequest)
  })
_sym_db.RegisterMessage(TrialListRequest)

TrialListEntry = _reflection.GeneratedProtocolMessageType('TrialListEntry', (_message.Message,), {
  'DESCRIPTOR' : _TRIALLISTENTRY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialListEntry)
  })
_sym_db.RegisterMessage(TrialListEntry)

TrialJoinRequest = _reflection.GeneratedProtocolMessageType('TrialJoinRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALJOINREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialJoinRequest)
  })
_sym_db.RegisterMessage(TrialJoinRequest)

TrialJoinReply = _reflection.GeneratedProtocolMessageType('TrialJoinReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALJOINREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialJoinReply)
  })
_sym_db.RegisterMessage(TrialJoinReply)

TrialHeartbeatRequest = _reflection.GeneratedProtocolMessageType('TrialHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALHEARTBEATREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialHeartbeatRequest)
  })
_sym_db.RegisterMessage(TrialHeartbeatRequest)

TrialHeartbeatReply = _reflection.GeneratedProtocolMessageType('TrialHeartbeatReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALHEARTBEATREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialHeartbeatReply)
  })
_sym_db.RegisterMessage(TrialHeartbeatReply)

TrialActionRequest = _reflection.GeneratedProtocolMessageType('TrialActionRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALACTIONREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialActionRequest)
  })
_sym_db.RegisterMessage(TrialActionRequest)

TrialActionReply = _reflection.GeneratedProtocolMessageType('TrialActionReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALACTIONREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialActionReply)
  })
_sym_db.RegisterMessage(TrialActionReply)

TrialRewardRequest = _reflection.GeneratedProtocolMessageType('TrialRewardRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALREWARDREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialRewardRequest)
  })
_sym_db.RegisterMessage(TrialRewardRequest)

TrialRewardReply = _reflection.GeneratedProtocolMessageType('TrialRewardReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALREWARDREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialRewardReply)
  })
_sym_db.RegisterMessage(TrialRewardReply)

TrialMessageRequest = _reflection.GeneratedProtocolMessageType('TrialMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALMESSAGEREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialMessageRequest)
  })
_sym_db.RegisterMessage(TrialMessageRequest)

TrialMessageReply = _reflection.GeneratedProtocolMessageType('TrialMessageReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALMESSAGEREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialMessageReply)
  })
_sym_db.RegisterMessage(TrialMessageReply)



_TRIALLIFECYCLE = _descriptor.ServiceDescriptor(
  name='TrialLifecycle',
  full_name='cogment.TrialLifecycle',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1355,
  serialized_end=1724,
  methods=[
  _descriptor.MethodDescriptor(
    name='StartTrial',
    full_name='cogment.TrialLifecycle.StartTrial',
    index=0,
    containing_service=None,
    input_type=_TRIALSTARTREQUEST,
    output_type=_TRIALSTARTREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TerminateTrial',
    full_name='cogment.TrialLifecycle.TerminateTrial',
    index=1,
    containing_service=None,
    input_type=_TERMINATETRIALREQUEST,
    output_type=_TERMINATETRIALREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTrialInfo',
    full_name='cogment.TrialLifecycle.GetTrialInfo',
    index=2,
    containing_service=None,
    input_type=_TRIALINFOREQUEST,
    output_type=_TRIALINFOREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='WatchTrials',
    full_name='cogment.TrialLifecycle.WatchTrials',
    index=3,
    containing_service=None,
    input_type=_TRIALLISTREQUEST,
    output_type=_TRIALLISTENTRY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.TrialLifecycle.Version',
    index=4,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRIALLIFECYCLE)

DESCRIPTOR.services_by_name['TrialLifecycle'] = _TRIALLIFECYCLE


_CLIENTACTOR = _descriptor.ServiceDescriptor(
  name='ClientActor',
  full_name='cogment.ClientActor',
  file=DESCRIPTOR,
  index=1,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1727,
  serialized_end=2169,
  methods=[
  _descriptor.MethodDescriptor(
    name='JoinTrial',
    full_name='cogment.ClientActor.JoinTrial',
    index=0,
    containing_service=None,
    input_type=_TRIALJOINREQUEST,
    output_type=_TRIALJOINREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ActionStream',
    full_name='cogment.ClientActor.ActionStream',
    index=1,
    containing_service=None,
    input_type=_TRIALACTIONREQUEST,
    output_type=_TRIALACTIONREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Heartbeat',
    full_name='cogment.ClientActor.Heartbeat',
    index=2,
    containing_service=None,
    input_type=_TRIALHEARTBEATREQUEST,
    output_type=_TRIALHEARTBEATREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendReward',
    full_name='cogment.ClientActor.SendReward',
    index=3,
    containing_service=None,
    input_type=_TRIALREWARDREQUEST,
    output_type=_TRIALREWARDREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendMessage',
    full_name='cogment.ClientActor.SendMessage',
    index=4,
    containing_service=None,
    input_type=_TRIALMESSAGEREQUEST,
    output_type=_TRIALMESSAGEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.ClientActor.Version',
    index=5,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_CLIENTACTOR)

DESCRIPTOR.services_by_name['ClientActor'] = _CLIENTACTOR

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from cogment.errors import InvalidRequestError
import cogment.api.environment_pb2_grpc as grpc_api

import cogment.api.environment_pb2 as env_api
import cogment.api.common_pb2 as common_api
from cogment.utils import list_versions
from cogment.session import RecvEvent, RecvMessage, RecvAction, EventType

from types import SimpleNamespace, ModuleType
import grpc.experimental.aio

from cogment.environment import _ServedEnvironmentSession, ENVIRONMENT_ACTOR_NAME

from cogment.trial import Trial

from prometheus_client import Summary, Counter

import atexit
import logging
import traceback
import typing
import asyncio
from time import time


def new_actions_table(settings, trial):
    actions_by_actor_class = settings.ActionsTable(trial)
    actions_by_actor_id = actions_by_actor_class.all_actions()

    return actions_by_actor_class, actions_by_actor_id


def pack_observations(env_session, observations, reply, tick_id=-1):
    timestamp = int(time() * 1000000000)

    new_obs = [None] * len(env_session._trial.actors)

    for tgt, obs in observations:
        for actor_index, actor in enumerate(env_session._trial.actors):
            if isinstance(tgt, int):
                new_obs[actor_index] = obs
            elif tgt == "*" or tgt == "*.*":
                new_obs[actor_index] = obs
            elif isinstance(tgt, str):
                if "." not in tgt:
                    if actor.name == tgt:
                        new_obs[actor_index] = obs
                else:
                    class_name = tgt.split(".")
                    if class_name[1] == actor.name:
                        new_obs[actor_index] = obs
                    elif class_name[1] == "*":
                        if actor.actor_class.name == class_name[0]:
                            new_obs[actor_index] = obs

    snapshots = [True] * len(env_session._trial.actors)

    for actor_index, actor in enumerate(env_session._trial.actors):
        if not new_obs[actor_index]:
            raise Exception(f"Actor [{actor.name}] is missing an observation")
        snapshots[actor_index] = isinstance(new_obs[actor_index], actor.actor_class.observation_space)

    # dupping time
    seen_observations = {}

    reply.observation_set.tick_id = tick_id
    reply.observation_set.timestamp = timestamp

    for actor_index, actor in enumerate(env_session._trial.actors):
        actor_obs = new_obs[actor_index]
        obs_id = id(actor_obs)
        obs_key = seen_observations.get(obs_id)
        if obs_key is None:
            obs_key = len(reply.observation_set.observations)

            obs_content = actor_obs.SerializeToString()
            observation_data = common_api.ObservationData(content=obs_content, snapshot=snapshots[actor_index])
            reply.observation_set.observations.append(observation_data)

            seen_observations[obs_id] = obs_key

        reply.observation_set.actors_map.append(obs_key)


def _process_reply(observations, env_session):
    rep = env_api.EnvActionReply()

    rep.rewards.extend(env_session._trial._gather_all_rewards())
    rep.messages.extend(env_session._trial._gather_all_messages(ENVIRONMENT_ACTOR_NAME))
    pack_observations(env_session, observations, rep)

    return rep


async def write_observations(context, env_session):
    try:
        while True:
            observations, final = await env_session._retrieve_obs()

            reply = _process_reply(observations, env_session)
            reply.final_update = final
            await context.write(reply)

            if final:
                env_session._trial.over = True
                env_session._new_event(RecvEvent(EventType.FINAL))
                break

    except asyncio.CancelledError:
        logging.debug("Environment 'write_observations' coroutine cancelled")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


def _process_actions(request, env_session):
    env_session._trial.tick_id = request.action_set.tick_id

    len_actions = len(request.action_set.actions)
    len_actors = len(env_session._trial._actions_by_actor_id)

    if len_actions != len_actors:
        raise Exception(f"Received {len_actions} actions but have {len_actors} actors")

    recv_event = RecvEvent(EventType.ACTIVE)
    for i, action in enumerate(env_session._trial._actions_by_actor_id):
        action.ParseFromString(request.action_set.actions[i])
        recv_event.actions.append(RecvAction(i, action))

    return recv_event


async def read_actions(context, env_session):
    try:
        while True:
            request = await context.read()

            if request == grpc.experimental.aio.EOF:
                logging.info(f"The orchestrator disconnected the environment")
                break

            env_session._new_event(_process_actions(request, env_session))

    except asyncio.CancelledError:
        logging.debug("Environment 'read_actions' coroutine cancelled")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise


class EnvironmentServicer(grpc_api.EnvironmentEndpointServicer):
    def __init__(self, env_impls, cog_settings, prometheus_registry):
        self.__impls = env_impls
        self.__env_sessions = {}
        self.__cog_settings = cog_settings

        self.update_count_per_trial = Summary(
            "environment_update_count_per_trial",
            "Number of update by trial",
            ["impl_name"],
            registry=prometheus_registry
        )
        self.trial_duration = Summary(
            "environment_trial_duration_in_second", "Trial duration", ["trial_actor"],
            registry=prometheus_registry
        )
        self.trials_started = Counter(
            "environment_trials_started", "Number of trial started", ["impl_name"],
            registry=prometheus_registry
        )
        self.trials_ended = Counter(
            "environment_trials_ended", "Number of trial ended", ["impl_name"],
            registry=prometheus_registry
        )
        self.messages_received = Counter(
            "environment_received_messages",
            "Number of messages received",
            ["impl_name"],
            registry=prometheus_registry
        )
        self.messages_sent = Counter(
            "environment_sent_messages",
            "Number of messages sent",
            ["impl_name"],
            registry=prometheus_registry
        )

        atexit.register(self.__cleanup)

        logging.info("Environment Service started")

    async def OnStart(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            if len(self.__impls) == 0:
                raise InvalidRequestError(message="No implementation registered", request=request)

            if request.impl_name:
                impl_name = request.impl_name
                impl = self.__impls.get(impl_name)
                if impl is None:
                    raise InvalidRequestError(message=f"Unknown environment impl [{impl_name}]", request=request)
            else:
                impl_name, impl = next(iter(self.__impls.items()))
                logging.info(f"Environment impl [{impl_name}] arbitrarily chosen")

            key = trial_id
            if key in self.__env_sessions:
                raise InvalidRequestError(message=f"Environment already exists for trial [{trial_id}]", request=request)

            self.trials_started.labels(impl_name).inc()

            trial = Trial(trial_id, request.actors_in_trial, self.__cog_settings)
            trial.tick_id = request.tick_id

            # action table time
            actions_by_actor_class, actions_by_actor_id = new_actions_table(self.__cog_settings, trial)

            trial._actions = actions_by_actor_class
            trial._actions_by_actor_id = actions_by_actor_id

            config = None
            if request.HasField("config"):
                if self.__cog_settings.environment.config_type is None:
                    raise Exception("Environment received config data of unknown type (was it defined in cogment.yaml)")
                config = self.__cog_settings.environment.config_type()
                config.ParseFromString(request.config.content)

            new_session = _ServedEnvironmentSession(impl.impl, trial, impl_name, config)
            self.__env_sessions[key] = new_session

            new_session._task = asyncio.create_task(new_session._run())

            observations, final = await new_session._retrieve_obs()
            if final:
                new_session._trial.over = True
                new_session._new_event(RecvEvent(EventType.FINAL))

            reply = env_api.EnvStartReply()
            pack_observations(new_session, observations, reply)

            return reply

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def OnAction(self, request_iterator, context):
        reader_task = None
        writer_task = None
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            key = trial_id
            env_session = self.__env_sessions[key]

            # We are going to have three concurrent coroutines:
            # - One that reads actions from the orchestrator
            # - One that sends observations to the orchestrator
            # - The environment's main (which will be the current coroutine)
            reader_task = asyncio.create_task(read_actions(context, env_session))
            writer_task = asyncio.create_task(write_observations(context, env_session))

            with self.trial_duration.labels(env_session.impl_name).time():
                await env_session._task

            self.update_count_per_trial.labels(env_session.impl_name).observe(env_session._trial.tick_id)

            if env_session.is_trial_over():
                logging.debug(f"User environment implementation for [{ENVIRONMENT_ACTOR_NAME}] returned")
            else:
                logging.info(f"User environment implementation for [{ENVIRONMENT_ACTOR_NAME}]"
                             f" running trial [{trial_id}] returned before end of trial")

            self.__env_sessions.pop(key, None)
            self.trials_ended.labels(env_session.impl_name).inc()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

        finally:
            if reader_task is not None:
                reader_task.cancel()
            if writer_task is not None:
                writer_task.cancel()

    async def OnMessage(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            key = trial_id
            env_session = self.__env_sessions[key]

            for message in request.messages:
                recv_event = RecvEvent(EventType.ACTIVE)
                recv_event.rewards = [RecvMessage(message)]
                env_session._new_event(recv_event)
                self.messages_received.labels(env_session.impl_name).inc()

            return env_api.EnvMessageReply()

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    async def OnEnd(self, request, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            key = trial_id
            env_session = self.__env_sessions[key]

            event = _process_actions(request, env_session)
            event.type = EventType.ENDING
            obs = await env_session._end_request(event)

            env_session._trial.over = True
            env_session._new_event(RecvEvent(EventType.FINAL))

            if obs:
                reply = _process_reply(obs, env_session)
            else:
                reply = env_api.EnvActionReply()
            reply.final_update = True

            self.__env_sessions.pop(key, None)
            self.trials_ended.labels(env_session.impl_name).inc()

            return reply

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise

    def __cleanup(self):
        self.__env_sessions.clear()
        atexit.unregister(self.__cleanup)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
from cogment.session import Session, EventType
from abc import ABC

ENVIRONMENT_ACTOR_NAME = "env"


class EnvironmentSession(Session):
    """This represents the environment being performed locally."""

    def __init__(self, impl, trial, impl_name, config):
        super().__init__(trial, "environment", impl, impl_name)
        self.config = config
        self._obs_queue = asyncio.Queue()
        self.__final_obs_future = None

    def start(self, observations):
        self._start()
        self._obs_queue.put_nowait((observations, False))

    def produce_observations(self, observations):
        if self._event_queue is None:
            logging.warning("Cannot send observations until the environment is started.")
            return

        if self.__final_obs_future is not None:
            self.__final_obs_future.set_result(observations)
        elif not self.is_trial_over():
            self._obs_queue.put_nowait((observations, False))
        else:
            logging.warning("Cannot send observation because the trial has ended.")

    def end(self, observations):
        if self.__final_obs_future is not None:
            self.__final_obs_future.set_result(observations)
        elif not self.is_trial_over():
            self._obs_queue.put_nowait((observations, True))
        else:
            logging.warning("Cannot end the environment because the trial has already ended.")

    async def _retrieve_obs(self):
        obs = await self._obs_queue.get()
        self._obs_queue.task_done()
        return obs

    async def _end_request(self, event):
        logging.debug("Environment received an end request")

        if self.is_trial_over():
            return None
        if self._event_queue is None:
            logging.warning("The environment received an end request before it was started.")
            return None

        self.__final_obs_future = asyncio.get_running_loop().create_future()

        await self._event_queue.put(event)

        result = await self.__final_obs_future
        self.__final_obs_future = None

        return result


class _ServedEnvironmentSession(EnvironmentSession):
    """An environment session that is served from an environment service."""

    def __init__(self, impl, trial, impl_name, config):
        super().__init__(impl, trial, impl_name, config)

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        assert self._trial is not None
        self._trial.add_reward(value, confidence, to, tick_id, user_data)

    def send_message(self, payload, to, to_environment=False):
        assert self._trial is not None
        self._trial.send_message(payload, to, to_environment)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.datalog_pb2_grpc as grpc_api
import cogment.api.datalog_pb2 as datalog_api
from cogment.trial import Trial
from cogment.datalog import _ServedDatalogSession
from cogment.utils import raw_params_to_user_params, list_versions
import logging
import traceback
import asyncio

import grpc.experimental.aio


async def read_sample(context, session):
    try:
        while True:
            request = await context.read()

            if request == grpc.experimental.aio.EOF:
                logging.info(f"The orchestrator disconnected from LogExpoterService.")
                break

            elif request.HasField("sample"):
                session._new_sample(request.sample)
            else:
                logging.warning(f"Invalid request received from the orchestrator : {request}")

    except asyncio.CancelledError:
        logging.debug(f"LogExporterService 'read_sample' coroutine cancelled.")

    except Exception:
        logging.error(f"{traceback.format_exc()}")
        raise

    # Exit the loop
    session._new_sample(None)


class LogExporterService(grpc_api.LogExporterServicer):

    def __init__(self, impl, cog_settings):
        self.__impl = impl
        self.__cog_settings = cog_settings
        logging.info("Log Exporter Service started")

    async def OnLogSample(self, request_iterator, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            request = await context.read()

            if not request.HasField("trial_params"):
                raise Exception(f"Initial logging request for [{trial_id}] does not contain parameters.")

            trial_params = raw_params_to_user_params(request.trial_params, self.__cog_settings)
            raw_trial_params = request.trial_params

            session = _ServedDatalogSession(self.__impl, trial_id, trial_params, raw_trial_params)
            session._task = asyncio.create_task(session._run())

            reader_task = asyncio.create_task(read_sample(context, session))

            # TODO: It is required to bypass a probable bug in easy_grpc that expects a stream to be "used".
            reply = datalog_api.LogExporterSampleReply()
            await context.write(reply)

            await session._task

        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise
        finally:
            if reader_task is not None:
                reader_task.cancel()

    async def Version(self, request, context):
        try:
            return list_versions()
        except Exception:
            logging.error(f"{traceback.format_exc()}")
            raise
